"""Class definition of the payload used to upload and send an artefact to ``hub``."""

from .artefact import CrossValidationReportArtefact, EstimatorReportArtefact

__all__ = [
    "EstimatorReportArtefact",
    "CrossValidationReportArtefact",
]



"""Class definition of the payload used to send a recall metric to ``hub``."""

from __future__ import annotations

from functools import cached_property
from typing import ClassVar, Literal

from pydantic import computed_field

from .metric import CrossValidationReportMetric, EstimatorReportMetric, cast_to_float


class Recall(EstimatorReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.recall"
    name: str = "recall"
    verbose_name: str = "Recall (macro)"
    greater_is_better: bool = True

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def value(self) -> float | None:  # noqa: D102
        try:
            function = self.report.metrics.recall
        except AttributeError:
            return None

        return cast_to_float(function(data_source=self.data_source, average="macro"))


class RecallTrain(Recall):  # noqa: D101
    data_source: Literal["train"] = "train"


class RecallTest(Recall):  # noqa: D101
    data_source: Literal["test"] = "test"


class RecallMean(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.recall"
    aggregate: ClassVar[Literal["mean"]] = "mean"
    name: str = "recall_mean"
    verbose_name: str = "Recall (macro) - MEAN"
    greater_is_better: bool = True

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def value(self) -> float | None:  # noqa: D102
        try:
            function = self.report.metrics.recall
        except AttributeError:
            return None

        dataframe = function(
            data_source=self.data_source, aggregate=self.aggregate, average="macro"
        )

        return cast_to_float(dataframe.iloc[0, 0])


class RecallTrainMean(RecallMean):  # noqa: D101
    data_source: Literal["train"] = "train"


class RecallTestMean(RecallMean):  # noqa: D101
    data_source: Literal["test"] = "test"


class RecallStd(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.recall"
    aggregate: ClassVar[Literal["std"]] = "std"
    name: str = "recall_std"
    verbose_name: str = "Recall (macro) - STD"
    greater_is_better: bool = False

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def value(self) -> float | None:  # noqa: D102
        try:
            function = self.report.metrics.recall
        except AttributeError:
            return None

        dataframe = function(
            data_source=self.data_source, aggregate=self.aggregate, average="macro"
        )

        return cast_to_float(dataframe.iloc[0, 0])


class RecallTrainStd(RecallStd):  # noqa: D101
    data_source: Literal["train"] = "train"


class RecallTestStd(RecallStd):  # noqa: D101
    data_source: Literal["test"] = "test"



"""Login to ``skore hub``."""

from __future__ import annotations

from contextlib import suppress
from webbrowser import open as open_webbrowser

from httpx import HTTPError
from rich.align import Align
from rich.panel import Panel

from .. import console
from ..client import api
from . import token as Token


def login(*, timeout=600):
    """Login to ``skore hub``."""
    with suppress(HTTPError):
        if Token.exists() and (Token.access() is not None):
            return

    url, device_code, user_code = api.get_oauth_device_login()

    if not open_webbrowser(url):
        console.print(
            Panel(
                Align(
                    f"Please visit this [link={url}]page[/link] to login.",
                    align="center",
                ),
                title="[cyan]Skore HUB[/cyan]",
                border_style="orange1",
                expand=False,
                padding=1,
                title_align="center",
            )
        )

    api.get_oauth_device_code_probe(device_code, timeout=timeout)
    api.post_oauth_device_callback(device_code, user_code)
    Token.save(*api.get_oauth_device_token(device_code))



"""Function definition of the artefact ``upload``."""

from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial
from math import ceil
from pathlib import Path
from typing import TYPE_CHECKING

from rich.progress import BarColumn, Progress, TextColumn, TimeElapsedColumn

from ..client.client import Client, HUBClient
from .serializer import Serializer

if TYPE_CHECKING:
    from typing import Any, Final

    import httpx

    from ..project.project import Project


SkinnedProgress = partial(
    Progress,
    TextColumn("[bold cyan blink]Uploading..."),
    BarColumn(
        complete_style="dark_orange",
        finished_style="dark_orange",
        pulse_style="orange1",
    ),
    TextColumn("[orange1]{task.percentage:>3.0f}%"),
    TimeElapsedColumn(),
    transient=True,
)


def upload_chunk(
    filepath: Path,
    client: httpx.Client,
    url: str,
    offset: int,
    length: int,
) -> str:
    """
    Upload a chunk of the serialized object to the artefacts storage.

    Parameters
    ----------
    filepath : ``Path``
        The path of the file containing the serialized object.
    client : ``httpx.Client``
        The client used to upload the chunk to the artefacts storage.
    url : str
        The url used to upload the chunk to the artefacts storage.
    offset : int
        The start of the chunk in the file containing the serialized object.
    length: int
        The length of the chunk in the file containing the serialized object.

    Returns
    -------
    etag : str
        The ETag assigned by the artefacts storage to the chunk, used to acknowledge the
        upload.

    Notes
    -----
    This function is in charge of reading its own chunk to reduce RAM footprint.
    """
    with filepath.open("rb") as file:
        file.seek(offset)

        response = client.put(
            url=url,
            content=file.read(length),
            headers={"Content-Type": "application/octet-stream"},
            timeout=30,
        )

        return response.headers["etag"]


# This is both the threshold at which an object is split into several small parts for
# upload, and the size of these small parts.
CHUNK_SIZE: Final[int] = int(1e7)  # ~10mb


def upload(project: Project, o: Any, type: str) -> str:
    """
    Upload an object to the artefacts storage.

    Parameters
    ----------
    project : ``Project``
        The project where to upload the object.
    o : Any
        The object to upload.
    type : str
        The type to associate to object in the artefacts storage.

    Returns
    -------
    checksum : str
        The checksum of the object after upload to the artefacts storage, based on its
        ``joblib`` serialization.

    Notes
    -----
    An object that was already uploaded in its whole will be ignored.
    """
    with (
        Serializer(o) as serializer,
        HUBClient() as hub_client,
        Client() as standard_client,
        ThreadPoolExecutor() as pool,
    ):
        # Ask for upload urls.
        response = hub_client.post(
            url=f"projects/{project.tenant}/{project.name}/artefacts",
            json=[
                {
                    "checksum": serializer.checksum,
                    "content_type": type,
                    "chunk_number": ceil(serializer.size / CHUNK_SIZE),
                }
            ],
        )

        # An empty response means that an artefact with the same checksum already
        # exists. The object doesn't have to be re-uploaded.
        if urls := response.json():
            task_to_chunk_id = {}

            # Upload each chunk of the serialized object to the artefacts storage, using
            # a disk temporary file.
            #
            # Each task is in charge of reading its own file chunk at runtime, to reduce
            # RAM footprint.
            #
            # Use `threading` over `asyncio` to ensure compatibility with Jupyter
            # notebooks, where the event loop is already running.
            for url in urls:
                chunk_id = url["chunk_id"] or 1
                task = pool.submit(
                    upload_chunk,
                    filepath=serializer.filepath,
                    client=standard_client,
                    url=url["upload_url"],
                    offset=((chunk_id - 1) * CHUNK_SIZE),
                    length=CHUNK_SIZE,
                )

                task_to_chunk_id[task] = chunk_id

            try:
                with SkinnedProgress() as progress:
                    tasks = as_completed(task_to_chunk_id)
                    total = len(task_to_chunk_id)
                    etags = dict(
                        sorted(
                            (
                                task_to_chunk_id[task],
                                task.result(),
                            )
                            for task in progress.track(tasks, total=total)
                        )
                    )
            except BaseException:
                # Cancel all remaining tasks, especially on `KeyboardInterrupt`.
                for task in task_to_chunk_id:
                    task.cancel()

                raise

            # Acknowledge the upload, to let the hub/storage rebuild the whole.
            hub_client.post(
                url=f"projects/{project.tenant}/{project.name}/artefacts/complete",
                json=[
                    {
                        "checksum": serializer.checksum,
                        "etags": etags,
                    }
                ],
            )

    return serializer.checksum



"""Class definition of the ``skore`` hub project."""

from __future__ import annotations

import itertools
import re
from functools import cached_property, wraps
from operator import itemgetter
from tempfile import TemporaryFile
from types import SimpleNamespace
from typing import TYPE_CHECKING

import joblib

from skore_hub_project.client.client import Client, HTTPStatusError, HUBClient
from skore_hub_project.protocol import CrossValidationReport, EstimatorReport

if TYPE_CHECKING:
    from typing import TypedDict

    class Metadata(TypedDict):  # noqa: D101
        id: str
        run_id: int
        key: str
        date: str
        learner: str
        ml_task: str
        report_type: str
        dataset: str
        rmse: float | None
        log_loss: float | None
        roc_auc: float | None
        fit_time: float | None
        predict_time: float | None
        rmse_mean: float | None
        log_loss_mean: float | None
        roc_auc_mean: float | None
        fit_time_mean: float | None
        predict_time_mean: float | None


def ensure_project_is_created(method):
    """
    Ensure project is created before executing any other operation.

    Notes
    -----
    This is based on the fact that requesting a ``run`` ID of a missing hub's project
    will trigger its creation on demand.
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        self.run_id  # noqa: B018
        return method(self, *args, **kwargs)

    return wrapper


class Project:
    """
    API to manage a collection of key-report pairs persisted in a hub storage.

    It communicates with the Probabl's ``skore hub`` storage, based on the pickle
    representation. Its constructor initializes a hub project by creating a new
    project or by loading an existing one from a defined tenant.

    The class main methods are :func:`~skore_hub_project.Project.put`,
    :func:`~skore_hub_project.reports.metadata` and
    :func:`~skore_hub_project.reports.get`, respectively to insert a key-report pair
    into the Project, to obtain the reports metadata and to get a specific report.

    Parameters
    ----------
    tenant : str
        The tenant of the project.

        A tenant is a ``skore hub`` concept that must be configured on the
        ``skore hub`` interface. It represents an isolated entity managing users,
        projects, and resources. It can be a company, organization, or team that
        operates independently within the system.
    name : str
        The name of the project.

    Attributes
    ----------
    tenant : str
        The tenant of the project.
    name : str
        The name of the project.
    run_id : int
        The current run identifier of the project.
    """

    __REPORT_URN_PATTERN = re.compile(
        r"skore:report:(?P<type>(estimator|cross-validation)):(?P<id>.+)"
    )

    def __init__(self, tenant: str, name: str):
        """
        Initialize a hub project.

        Initialize a hub project by creating a new project or by loading an existing
        one from a defined tenant.

        Parameters
        ----------
        tenant : Path
            The tenant of the project.

            A tenant is a ``skore hub`` concept that must be configured on the
            ``skore hub`` interface. It represents an isolated entity managing users,
            projects, and resources. It can be a company, organization, or team that
            operates independently within the system.
        name : str
            The name of the project.
        """
        self.__tenant = tenant
        self.__name = name

    @property
    def tenant(self) -> str:
        """The tenant of the project."""
        return self.__tenant

    @property
    def name(self) -> str:
        """The name of the project."""
        return self.__name

    @cached_property
    def run_id(self) -> int:
        """The current run identifier of the project."""
        with HUBClient() as client:
            request = client.post(f"projects/{self.tenant}/{self.name}/runs")
            run = request.json()

        return run["id"]

    @ensure_project_is_created
    def put(self, key: str, report: EstimatorReport | CrossValidationReport):
        """
        Put a key-report pair to the hub project.

        If the key already exists, its last report is modified to point to this new
        report, while keeping track of the report history.

        Parameters
        ----------
        key : str
            The key to associate with ``report`` in the hub project.
        report : skore.EstimatorReport | skore.CrossValidationReport
            The report to associate with ``key`` in the hub project.

        Raises
        ------
        TypeError
            If the combination of parameters are not valid.
        """
        from ..report import CrossValidationReportPayload, EstimatorReportPayload

        if not isinstance(key, str):
            raise TypeError(f"Key must be a string (found '{type(key)}')")

        Payload: type

        if isinstance(report, EstimatorReport):
            Payload = EstimatorReportPayload
            url = f"projects/{self.tenant}/{self.name}/estimator-reports"
        elif isinstance(report, CrossValidationReport):
            Payload = CrossValidationReportPayload
            url = f"projects/{self.tenant}/{self.name}/cross-validation-reports"
        else:
            raise TypeError(
                f"Report must be a `skore.EstimatorReport` or "
                f"`skore.CrossValidationReport` (found '{type(report)}')"
            )

        payload = Payload(project=self, key=key, report=report)
        payload_dict = payload.model_dump()

        with HUBClient() as client:
            client.post(url=url, json=payload_dict)

    @ensure_project_is_created
    def get(self, urn: str) -> EstimatorReport | CrossValidationReport:
        """Get a persisted report by its URN."""
        if m := re.match(Project.__REPORT_URN_PATTERN, urn):
            url = f"projects/{self.tenant}/{self.name}/{m['type']}-reports/{m['id']}"
        else:
            raise ValueError(
                f"URN '{urn}' format does not match '{Project.__REPORT_URN_PATTERN}'"
            )

        # Retrieve report metadata.
        with HUBClient() as client:
            response = client.get(url=url)

        metadata = response.json()
        checksum = metadata["raw"]["checksum"]

        # Ask for read url.
        with HUBClient() as client:
            response = client.get(
                url=f"projects/{self.tenant}/{self.name}/artefacts/read",
                params={"artefact_checksum": [checksum]},
            )

        url = response.json()[0]["url"]

        # Download pickled report before unpickling it.
        #
        # It uses streaming responses that do not load the entire response body into
        # memory at once.
        with (
            TemporaryFile(mode="w+b") as tmpfile,
            Client() as client,
            client.stream(method="GET", url=url, timeout=30) as response,
        ):
            for data in response.iter_bytes():
                tmpfile.write(data)

            tmpfile.seek(0)

            return joblib.load(tmpfile)

    @ensure_project_is_created
    def summarize(self) -> list[Metadata]:
        """Obtain metadata/metrics for all persisted reports in insertion order."""

        def dto(response):
            report_type, summary = response
            metrics = {
                metric["name"]: metric["value"]
                for metric in summary["metrics"]
                if metric["data_source"] in (None, "test")
            }

            return {
                "id": summary["urn"],
                "run_id": summary["run_id"],
                "key": summary["key"],
                "date": summary["created_at"],
                "learner": summary["estimator_class_name"],
                "ml_task": summary["ml_task"],
                "report_type": report_type,
                "dataset": summary["dataset_fingerprint"],
                "rmse": metrics.get("rmse"),
                "log_loss": metrics.get("log_loss"),
                "roc_auc": metrics.get("roc_auc"),
                "fit_time": metrics.get("fit_time"),
                "predict_time": metrics.get("predict_time"),
                "rmse_mean": metrics.get("rmse_mean"),
                "log_loss_mean": metrics.get("log_loss_mean"),
                "roc_auc_mean": metrics.get("roc_auc_mean"),
                "fit_time_mean": metrics.get("fit_time_mean"),
                "predict_time_mean": metrics.get("predict_time_mean"),
            }

        with HUBClient() as client:
            responses = itertools.chain(
                zip(
                    itertools.repeat("estimator"),
                    client.get(
                        f"projects/{self.tenant}/{self.name}/estimator-reports/"
                    ).json(),
                ),
                zip(
                    itertools.repeat("cross-validation"),
                    client.get(
                        f"projects/{self.tenant}/{self.name}/cross-validation-reports/"
                    ).json(),
                ),
            )

        return sorted(map(dto, responses), key=itemgetter("date"))

    @property
    @ensure_project_is_created
    def reports(self):
        """Accessor for interaction with the persisted reports."""

        def get(urn: str) -> EstimatorReport | CrossValidationReport:
            """
            Get a persisted report by its URN.

            .. deprecated
              The ``Project.reports.get`` function will be removed in favor of
              ``Project.get`` in a near future.
            """
            return self.get(urn)

        def metadata() -> list[Metadata]:
            """
            Obtain metadata/metrics for all persisted reports in insertion order.

            .. deprecated
              The ``Project.reports.metadata`` function will be removed in favor of
              ``Project.summarize`` in a near future.
            """
            return self.summarize()

        return SimpleNamespace(get=get, metadata=metadata)

    def __repr__(self) -> str:  # noqa: D105
        return f"Project(mode='hub', name='{self.name}', tenant='{self.tenant}')"

    @staticmethod
    def delete(tenant: str, name: str):
        """
        Delete a hub project.

        Parameters
        ----------
        tenant : Path
            The tenant of the project.

            A tenant is a ``skore hub`` concept that must be configured on the
            ``skore hub`` interface. It represents an isolated entity managing users,
            projects, and resources. It can be a company, organization, or team that
            operates independently within the system.
        name : str
            The name of the project.
        """
        with HUBClient() as client:
            try:
                client.delete(f"projects/{tenant}/{name}")
            except HTTPStatusError as e:
                if e.response.status_code == 403:
                    raise PermissionError(
                        f"Failed to delete the project; "
                        f"please contact the '{tenant}' owner"
                    ) from e
                raise



"""Class definition of the payload used to send a report to ``hub``."""

from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import cached_property
from typing import ClassVar, cast

from pydantic import BaseModel, ConfigDict, Field, computed_field

from skore_hub_project import Project
from skore_hub_project.artefact.artefact import Artefact
from skore_hub_project.media.media import Media
from skore_hub_project.metric.metric import Metric
from skore_hub_project.protocol import CrossValidationReport, EstimatorReport


class ReportPayload(ABC, BaseModel):
    """
    Payload used to send a report to ``hub``.

    Attributes
    ----------
    METRICS : ClassVar[tuple[Metric, ...]]
        The metric classes that have to be computed from the report.
    MEDIAS : ClassVar[tuple[Media, ...]]
        The media classes that have to be computed from the report.
    project : Project
        The project to which the report payload should be sent.
    report : EstimatorReport | CrossValidationReport
        The report on which to calculate the payload to be sent.
    upload : bool, optional
        Upload the report to the artefacts storage, default True.
    key : str
        The key to associate to the report.
    """

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    METRICS: ClassVar[tuple[Metric, ...]]
    MEDIAS: ClassVar[tuple[Media, ...]]

    project: Project = Field(repr=False, exclude=True)
    report: EstimatorReport | CrossValidationReport = Field(repr=False, exclude=True)
    upload: bool = Field(default=True, repr=False, exclude=True)
    key: str

    @computed_field  # type: ignore[prop-decorator]
    @property
    def run_id(self) -> int:
        """The current run identifier of the project."""
        return self.project.run_id

    @computed_field  # type: ignore[prop-decorator]
    @property
    def estimator_class_name(self) -> str:
        """The name of the report's estimator."""
        return self.report.estimator_name_

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def dataset_fingerprint(self) -> str:
        """The hash of the targets in the test-set."""
        import joblib

        return joblib.hash(
            self.report.y_test if hasattr(self.report, "y_test") else self.report.y
        )

    @computed_field  # type: ignore[prop-decorator]
    @property
    def ml_task(self) -> str:
        """The type of ML task covered by the report."""
        return self.report.ml_task

    @computed_field  # type: ignore[prop-decorator]
    @property
    @abstractmethod
    def parameters(self) -> Artefact | dict[()]:
        """
        The checksum of the instance.

        The checksum of the instance that was assigned after being uploaded to the
        artefact storage. It is based on its ``joblib`` serialization and mainly used to
        retrieve it from the artefacts storage.

        .. deprecated
          The ``parameters`` property will be removed in favor of a new ``checksum``
          property in a near future.
        """

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def metrics(self) -> list[Metric]:
        """
        The list of scalar metrics that have been computed from the report.

        Notes
        -----
        Unavailable metrics have been filtered out.

        All metrics whose value is not a scalar are currently ignored:
        - ignore ``list[float]`` for multi-output ML task,
        - ignore ``dict[str: float]`` for multi-classes ML task.

        The position field is used to drive the ``hub``'s parallel coordinates plot:
        - int [0, inf[, to be displayed at the position,
        - None, not to be displayed.
        """
        payloads = [
            payload
            for metric in self.METRICS
            if (payload := metric(report=self.report)).value is not None
        ]

        return payloads

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def related_items(self) -> list[Media]:
        """
        The list of medias that have been computed from the report.

        Medias are `pandas.Dataframe`, SVG images, Python dictionaries or HTML string.

        Notes
        -----
        Unavailable medias have been filtered out.
        """
        payloads = [
            payload
            for media in cast(list[Callable], self.MEDIAS)
            if (payload := media(report=self.report)).representation is not None
        ]

        return payloads



"""Class definition of the payload used to send a model category media to ``hub``."""

from functools import cached_property
from typing import Literal

from pydantic import Field, computed_field

from skore_hub_project.protocol import CrossValidationReport, EstimatorReport

from .media import Media, Representation


class EstimatorHtmlRepr(Media):  # noqa: D101
    report: EstimatorReport | CrossValidationReport = Field(repr=False, exclude=True)
    key: str = "estimator_html_repr"
    verbose_name: str = "Estimator HTML representation"
    category: Literal["model"] = "model"

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def representation(self) -> Representation:  # noqa: D102
        import sklearn.utils

        return Representation(
            media_type="text/html",
            value=sklearn.utils.estimator_html_repr(self.report.estimator),
        )



"""Class definitions of the payloads used to send a report to ``hub``."""

from .cross_validation_report import CrossValidationReportPayload
from .estimator_report import EstimatorReportPayload

__all__ = [
    "CrossValidationReportPayload",
    "EstimatorReportPayload",
]



"""Client exchanging with ``skore hub``."""

from __future__ import annotations

import logging
from contextlib import suppress
from http import HTTPStatus
from os import environ
from time import sleep
from typing import Final
from urllib.parse import urljoin

from httpx import (
    URL,
    Headers,
    HTTPError,
    HTTPStatusError,
    NetworkError,
    RemoteProtocolError,
    Response,
    TimeoutException,
)
from httpx import (
    Client as HTTPXClient,
)
from httpx._types import HeaderTypes

from ..authentication import token as Token

logger = logging.getLogger(__name__)


class Client(HTTPXClient):
    """
    Client with a retry strategy based on exponential backoff algorithm.

    Parameters
    ----------
    retry : bool, optional
        Retry request on fails, default True.
    retry_total: int | None, optional
        Total number of retries to allow, default 10.
        Set to None to remove this constraint.
    retry_backoff_factor : float, optional
        A backoff factor to apply between retries, default 0.25.
    retry_backoff_max : float, optional
        Maximum backoff time, default 120.
    """

    RETRYABLE_STATUS_CODES: Final[frozenset[HTTPStatus]] = frozenset(
        (
            HTTPStatus.REQUEST_TIMEOUT,
            HTTPStatus.TOO_EARLY,
            HTTPStatus.TOO_MANY_REQUESTS,
            HTTPStatus.BAD_GATEWAY,
            HTTPStatus.SERVICE_UNAVAILABLE,
            HTTPStatus.GATEWAY_TIMEOUT,
        )
    )

    RETRYABLE_EXCEPTIONS: Final[tuple[HTTPError, ...]] = (
        TimeoutException,
        NetworkError,
        RemoteProtocolError,
    )

    STATUS_CLASS_TO_ERROR_TYPE: Final[dict[int, str]] = {
        1: "Informational response",
        3: "Redirect response",
        4: "Client error",
        5: "Server error",
    }

    def __init__(
        self,
        *,
        retry: bool = True,
        retry_total: int | None = 10,
        retry_backoff_factor: float = 0.25,
        retry_backoff_max: float = 120,
    ):
        super().__init__(follow_redirects=True, timeout=30)

        self.retry = retry
        self.retry_total = retry_total if retry_total is not None else float("inf")
        self.retry_backoff_factor = retry_backoff_factor
        self.retry_backoff_max = retry_backoff_max

    def request(self, *args, **kwargs) -> Response:
        """Execute request with retry strategy."""
        retries = 0

        while True:
            try:
                response = super().request(*args, **kwargs)
            except self.RETRYABLE_EXCEPTIONS:
                if not self.retry:
                    raise
            else:
                if response.is_success:
                    return response

                overlimit = retries >= self.retry_total
                unretryable = response.status_code not in self.RETRYABLE_STATUS_CODES

                if (not self.retry) or overlimit or unretryable:
                    break

            timeout = self.retry_backoff_factor * (2**retries)
            retries += 1

            sleep_duration = min(timeout, self.retry_backoff_max)

            logger.warn(
                "Request failed to reach server. "
                f"Trying again in {sleep_duration} seconds."
            )

            sleep(sleep_duration)

        # Raise extended exception with body message when available.
        status_class = response.status_code // 100
        error_type = self.STATUS_CLASS_TO_ERROR_TYPE.get(
            status_class, "Invalid status code"
        )

        message = (
            f"{error_type} '{response.status_code} {response.reason_phrase}' "
            f"for url '{response.url}'"
        )

        with suppress(Exception):
            message += f": {response.json()['message']}"

        raise HTTPStatusError(message, request=response.request, response=response)


class AuthenticationError(Exception):
    """An exception dedicated to authentication."""


class HUBClient(Client):
    """
    Client exchanging with ``skore hub``.

    Parameters
    ----------
    authenticated : bool, optional
        Use headers with API key or token, default True.
    """

    URI: Final[str] = environ.get("SKORE_HUB_URI", "https://api.skore.probabl.ai")

    def __init__(
        self,
        *,
        authenticated=True,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.authenticated = authenticated

    def request(
        self,
        method: str,
        url: URL | str,
        headers: HeaderTypes | None = None,
        **kwargs,
    ) -> Response:
        """Execute request with authorization."""
        headers = Headers(headers)

        # Overload headers with our custom headers (API key or token)
        if self.authenticated:
            if (apikey := environ.get("SKORE_HUB_API_KEY")) is not None:
                headers.update({"X-API-Key": f"{apikey}"})
            else:
                if not Token.exists():
                    raise AuthenticationError(
                        "You are not logged in. Please run `skore-hub-login`"
                    )

                headers.update({"Authorization": f"Bearer {Token.access()}"})

        # Prefix ``url`` by the hub URI when it's possible
        url = urljoin(self.URI, str(url))

        return super().request(method=method, url=url, headers=headers, **kwargs)



"""Class definition of the payload used to send a metric to ``hub``."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from contextlib import suppress
from functools import cached_property, reduce
from math import isfinite
from typing import Any, ClassVar, Literal, cast

from pydantic import BaseModel, ConfigDict, Field, computed_field

from skore_hub_project.protocol import CrossValidationReport, EstimatorReport


def cast_to_float(value: Any) -> float | None:
    """Cast value to float."""
    with suppress(TypeError):
        if isfinite(value := float(value)):
            return value

    return None


class Metric(ABC, BaseModel):
    """
    Payload used to send a metric to ``hub``.

    Attributes
    ----------
    name : str
        Name of the metric.
    verbose_name : str
        Verbose name of the metric.
    data_source : Literal["train", "test"] | None, optional
        Data source of the metric when it can be declined in several ways, default None.
    greater_is_better: bool | None, optional
        Indicator of "greater value is better", default None.
    position: int | None, optional
        Indicator of the "position" of the metric in the parallel coordinates plot,
        default None to disable its display.
    """

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    name: str
    verbose_name: str
    data_source: Literal["train", "test"] | None = None
    greater_is_better: bool | None = None
    position: int | None = None

    @computed_field  # type: ignore[prop-decorator]
    @property
    @abstractmethod
    def value(self) -> float | None:
        """The value of the metric."""


class EstimatorReportMetric(Metric):
    """
    Payload used to send an estimator report metric.

    Attributes
    ----------
    name : str
        Name of the metric.
    verbose_name : str
        Verbose name of the metric.
    data_source : Literal["train", "test"] | None, optional
        Data source of the metric when it can be declined in several ways, default None.
    greater_is_better: bool | None, optional
        Indicator of "greater value is better", default None.
    position: int | None, optional
        Indicator of the "position" of the metric in the parallel coordinates plot,
        default None to disable its display.
    report: EstimatorReport
        The report on which compute the metric.
    accessor : ClassVar[str]
        The "accessor" of the metric i.e., the path to the metric calculation function.
    """

    report: EstimatorReport = Field(repr=False, exclude=True)
    accessor: ClassVar[str]

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def value(self) -> float | None:
        """The value of the metric."""
        try:
            function = cast(
                Callable,
                reduce(getattr, self.accessor.split("."), self.report),
            )
        except AttributeError:
            return None

        return cast_to_float(function(data_source=self.data_source))


class CrossValidationReportMetric(Metric):
    """
    Payload used to send a cross-validation report metric, usually MEAN or STD.

    Attributes
    ----------
    name : str
        Name of the metric.
    verbose_name : str
        Verbose name of the metric.
    data_source : Literal["train", "test"] | None, optional
        Data source of the metric when it can be declined in several ways, default None.
    greater_is_better: bool | None, optional
        Indicator of "greater value is better", default None.
    position: int | None, optional
        Indicator of the "position" of the metric in the parallel coordinates plot,
        default None to disable its display.
    report: CrossValidationReport
        The report on which compute the metric.
    accessor : ClassVar[str]
        The "accessor" of the metric i.e., the path to the metric calculation function.
    aggregate : ClassVar[Literal["mean", "std"]]
        The aggregation parameter passed to the ``accessor``.
    """

    report: CrossValidationReport = Field(repr=False, exclude=True)
    accessor: ClassVar[str]
    aggregate: ClassVar[Literal["mean", "std"]]

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def value(self) -> float | None:
        """The value of the metric."""
        try:
            function = cast(
                Callable,
                reduce(getattr, self.accessor.split("."), self.report),
            )
        except AttributeError:
            return None

        dataframe = function(data_source=self.data_source, aggregate=self.aggregate)

        return cast_to_float(dataframe.iloc[0, 0])



"""Class definition of the payload used to send a ROC AUC metric to ``hub``."""

from __future__ import annotations

from typing import ClassVar, Literal

from .metric import CrossValidationReportMetric, EstimatorReportMetric


class RocAuc(EstimatorReportMetric):  # noqa: D101
    accessor: ClassVar[Literal["metrics.roc_auc"]] = "metrics.roc_auc"
    name: Literal["roc_auc"] = "roc_auc"
    verbose_name: Literal["ROC AUC"] = "ROC AUC"
    greater_is_better: Literal[True] = True
    position: Literal[3] = 3


class RocAucTrain(RocAuc):  # noqa: D101
    data_source: Literal["train"] = "train"


class RocAucTest(RocAuc):  # noqa: D101
    data_source: Literal["test"] = "test"


class RocAucMean(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[Literal["metrics.roc_auc"]] = "metrics.roc_auc"
    aggregate: ClassVar[Literal["mean"]] = "mean"
    name: Literal["roc_auc_mean"] = "roc_auc_mean"
    verbose_name: Literal["ROC AUC - MEAN"] = "ROC AUC - MEAN"
    greater_is_better: Literal[True] = True
    position: Literal[3] = 3


class RocAucTrainMean(RocAucMean):  # noqa: D101
    data_source: Literal["train"] = "train"


class RocAucTestMean(RocAucMean):  # noqa: D101
    data_source: Literal["test"] = "test"


class RocAucStd(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[Literal["metrics.roc_auc"]] = "metrics.roc_auc"
    aggregate: ClassVar[Literal["std"]] = "std"
    name: Literal["roc_auc_std"] = "roc_auc_std"
    verbose_name: Literal["ROC AUC - STD"] = "ROC AUC - STD"
    greater_is_better: Literal[False] = False


class RocAucTrainStd(RocAucStd):  # noqa: D101
    data_source: Literal["train"] = "train"


class RocAucTestStd(RocAucStd):  # noqa: D101
    data_source: Literal["test"] = "test"



"""Class definition of the payload used to send an estimator report to ``hub``."""

from functools import cached_property
from typing import ClassVar, cast

from pydantic import Field, computed_field

from skore_hub_project.artefact import EstimatorReportArtefact
from skore_hub_project.media import (
    Coefficients,
    EstimatorHtmlRepr,
    MeanDecreaseImpurity,
    PermutationTest,
    PermutationTrain,
    PrecisionRecallTest,
    PrecisionRecallTrain,
    PredictionErrorTest,
    PredictionErrorTrain,
    RocTest,
    RocTrain,
    TableReportTest,
    TableReportTrain,
)
from skore_hub_project.media.media import Media
from skore_hub_project.metric import (
    AccuracyTest,
    AccuracyTrain,
    BrierScoreTest,
    BrierScoreTrain,
    FitTime,
    LogLossTest,
    LogLossTrain,
    PrecisionTest,
    PrecisionTrain,
    PredictTimeTest,
    PredictTimeTrain,
    R2Test,
    R2Train,
    RecallTest,
    RecallTrain,
    RmseTest,
    RmseTrain,
    RocAucTest,
    RocAucTrain,
)
from skore_hub_project.metric.metric import Metric
from skore_hub_project.protocol import EstimatorReport
from skore_hub_project.report.report import ReportPayload


class EstimatorReportPayload(ReportPayload):
    """
    Payload used to send an estimator report to ``hub``.

    Attributes
    ----------
    METRICS : ClassVar[tuple[Metric, ...]]
        The metric classes that have to be computed from the report.
    MEDIAS : ClassVar[tuple[Media, ...]]
        The media classes that have to be computed from the report.
    project : Project
        The project to which the report payload should be sent.
    report : EstimatorReport
        The report on which to calculate the payload to be sent.
    upload : bool, optional
        Upload the report to the artefacts storage, default True.
    key : str
        The key to associate to the report.
    """

    METRICS: ClassVar[tuple[Metric, ...]] = cast(
        tuple[Metric, ...],
        (
            AccuracyTest,
            AccuracyTrain,
            BrierScoreTest,
            BrierScoreTrain,
            LogLossTest,
            LogLossTrain,
            PrecisionTest,
            PrecisionTrain,
            R2Test,
            R2Train,
            RecallTest,
            RecallTrain,
            RmseTest,
            RmseTrain,
            RocAucTest,
            RocAucTrain,
            # timings must be calculated last
            FitTime,
            PredictTimeTest,
            PredictTimeTrain,
        ),
    )
    MEDIAS: ClassVar[tuple[Media, ...]] = cast(
        tuple[Media, ...],
        (
            Coefficients,
            EstimatorHtmlRepr,
            MeanDecreaseImpurity,
            PermutationTest,
            PermutationTrain,
            PrecisionRecallTest,
            PrecisionRecallTrain,
            PredictionErrorTest,
            PredictionErrorTrain,
            RocTest,
            RocTrain,
            TableReportTest,
            TableReportTrain,
        ),
    )

    report: EstimatorReport = Field(repr=False, exclude=True)

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def parameters(self) -> EstimatorReportArtefact | dict:
        """
        The checksum of the instance.

        The checksum of the instance that was assigned after being uploaded to the
        artefact storage. It is based on its ``joblib`` serialization and mainly used to
        retrieve it from the artefacts storage.

        .. deprecated
          The ``parameters`` property will be removed in favor of a new ``checksum``
          property in a near future.
        """
        if self.upload:
            return EstimatorReportArtefact(project=self.project, report=self.report)
        return {}



"""Class definition of the payload used to send feature importance medias to ``hub``."""

from collections.abc import Callable
from functools import cached_property, reduce
from inspect import signature
from typing import ClassVar, Literal, cast

from pandas import DataFrame
from pydantic import Field, computed_field

from skore_hub_project.protocol import CrossValidationReport, EstimatorReport

from .media import Media, Representation


class FeatureImportance(Media):  # noqa: D101
    report: EstimatorReport | CrossValidationReport = Field(repr=False, exclude=True)
    accessor: ClassVar[str]
    category: Literal["feature_importance"] = "feature_importance"

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def representation(self) -> Representation | None:  # noqa: D102
        try:
            function = cast(
                Callable,
                reduce(getattr, self.accessor.split("."), self.report),
            )
        except AttributeError:
            return None

        function_parameters = signature(function).parameters
        function_kwargs = {
            k: v for k, v in self.attributes.items() if k in function_parameters
        }

        result = function(**function_kwargs)

        if not isinstance(result, DataFrame):
            result = result.frame()

        serialized = result.fillna("NaN").to_dict(orient="tight")

        return Representation(media_type="application/vnd.dataframe", value=serialized)


class Permutation(FeatureImportance):  # noqa: D101
    report: EstimatorReport = Field(repr=False, exclude=True)
    accessor: ClassVar[str] = "feature_importance.permutation"
    key: str = "permutation"
    verbose_name: str = "Feature importance - Permutation"


class PermutationTrain(Permutation):  # noqa: D101
    attributes: dict = {"data_source": "train", "method": "permutation"}


class PermutationTest(Permutation):  # noqa: D101
    attributes: dict = {"data_source": "test", "method": "permutation"}


class MeanDecreaseImpurity(FeatureImportance):  # noqa: D101
    report: EstimatorReport = Field(repr=False, exclude=True)
    accessor: ClassVar[str] = "feature_importance.mean_decrease_impurity"
    key: str = "mean_decrease_impurity"
    verbose_name: str = "Feature importance - Mean Decrease Impurity (MDI)"
    attributes: dict = {"method": "mean_decrease_impurity"}


class Coefficients(FeatureImportance):  # noqa: D101
    accessor: ClassVar[str] = "feature_importance.coefficients"
    key: str = "coefficients"
    verbose_name: str = "Feature importance - Coefficients"
    attributes: dict = {"method": "coefficients"}



"""API used to exchange with ``skore hub``."""

from datetime import datetime
from time import sleep

from httpx import HTTPError, TimeoutException


def get_oauth_device_login(success_uri: str | None = None):
    """Initiate device OAuth flow.

    Initiates the OAuth device flow.
    Provides the user with a URL and a OTP code to authenticate the device.

    Parameters
    ----------
    success_uri : str, optional
        The URI to redirect to after successful authentication.
        If not provided, defaults to None.

    Returns
    -------
    tuple
        A tuple containing:
        - authorization_url: str
            The URL to which the user needs to navigate
        - device_code: str
            The device code used for authentication
        - user_code: str
            The user code that needs to be entered on the authorization page
    """
    from skore_hub_project.client.client import HUBClient

    url = "identity/oauth/device/login"
    params = {"success_uri": success_uri} if success_uri is not None else {}

    with HUBClient(authenticated=False) as client:
        response = client.get(url, params=params).json()

        return (
            response["authorization_url"],
            response["device_code"],
            response["user_code"],
        )


def post_oauth_device_callback(state: str, user_code: str):
    """Validate the user-provided device code.

    This endpoint verifies the code entered by the user during the device auth flow.

    Parameters
    ----------
    state: str
        The unique value identifying the device flow.
    user_code: str
        The code entered by the user.
    """
    from skore_hub_project.client.client import HUBClient

    url = "identity/oauth/device/callback"
    data = {"state": state, "user_code": user_code}

    with HUBClient(authenticated=False) as client:
        client.post(url, data=data)


def get_oauth_device_token(device_code: str):
    """Exchanges the device code for an access token.

    This endpoint completes the device authorization flow
    by exchanging the validated device code for an access token.

    Parameters
    ----------
    device_code : str
        The device code to exchange for tokens

    Returns
    -------
    tuple
        A tuple containing:
        - access_token : str
            The OAuth access token
        - refresh_token : str
            The OAuth refresh token
        - expires_at : str
            The expiration datetime as ISO 8601 str of the access token
    """
    from skore_hub_project.client.client import HUBClient

    url = "identity/oauth/device/token"
    params = {"device_code": device_code}

    with HUBClient(authenticated=False) as client:
        response = client.get(url, params=params).json()
        tokens = response["token"]

        return (
            tokens["access_token"],
            tokens["refresh_token"],
            tokens["expires_at"],
        )


def get_oauth_device_code_probe(device_code: str, *, timeout=600):
    """Ensure authorization code is acknowledged.

    Start polling, wait for the authorization code to be acknowledged by the hub.
    This is mandatory to be authorize to exchange with a token.

    Parameters
    ----------
    device_code : str
        The device code to exchange for tokens.
    """
    from skore_hub_project.client.client import HUBClient

    url = "identity/oauth/device/code-probe"
    params = {"device_code": device_code}

    with HUBClient(authenticated=False) as client:
        start = datetime.now()

        while True:
            try:
                client.get(url, params=params)
            except HTTPError as exc:
                if exc.response.status_code != 400:
                    raise

                if (datetime.now() - start).total_seconds() >= timeout:
                    raise TimeoutException("Authentication timeout") from exc

                sleep(0.5)
            else:
                break


def post_oauth_refresh_token(refresh_token: str):
    """Refresh an access token using a provided refresh token.

    This endpoint allows a client to obtain a new access token
    by providing a valid refresh token.

    Parameters
    ----------
    refresh_token : str
        A valid refresh token

    Returns
    -------
    tuple
        A tuple containing:
        - access_token : str
            The OAuth access token
        - refresh_token : str
            The OAuth refresh token
        - expires_at : str
            The expiration datetime as ISO 8601 str of the access token
    """
    from skore_hub_project.client.client import HUBClient

    url = "identity/oauth/token/refresh"
    json = {"refresh_token": refresh_token}

    with HUBClient(authenticated=False) as client:
        response = client.post(url, json=json).json()

        return (
            response["access_token"],
            response["refresh_token"],
            response["expires_at"],
        )



"""Class definition of the payload used to send a performance media to ``hub``."""

from collections.abc import Callable
from functools import cached_property, reduce
from inspect import signature
from io import BytesIO
from typing import ClassVar, Literal, cast

from matplotlib import pyplot as plt
from pydantic import Field, computed_field

from skore_hub_project.protocol import CrossValidationReport, EstimatorReport

from .media import Media, Representation


class Performance(Media):  # noqa: D101
    report: EstimatorReport | CrossValidationReport = Field(repr=False, exclude=True)
    accessor: ClassVar[str]
    category: Literal["performance"] = "performance"

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def representation(self) -> Representation | None:  # noqa: D102
        from skore_hub_project import bytes_to_b64_str, switch_mpl_backend

        try:
            function = cast(
                Callable,
                reduce(getattr, self.accessor.split("."), self.report),
            )
        except AttributeError:
            return None

        function_parameters = signature(function).parameters
        function_kwargs = {
            k: v for k, v in self.attributes.items() if k in function_parameters
        }

        display = function(**function_kwargs)

        with switch_mpl_backend(), BytesIO() as stream:
            display.plot()
            display.figure_.savefig(stream, format="svg", bbox_inches="tight")

            figure_bytes = stream.getvalue()
            figure_b64_str = bytes_to_b64_str(figure_bytes)

            plt.close(display.figure_)

        return Representation(
            media_type="image/svg+xml;base64",
            value=figure_b64_str,
        )


class PrecisionRecall(Performance):  # noqa: D101
    accessor: ClassVar[str] = "metrics.precision_recall"
    key: str = "precision_recall"
    verbose_name: str = "Precision Recall"


class PrecisionRecallTrain(PrecisionRecall):  # noqa: D101
    attributes: dict = {"data_source": "train"}


class PrecisionRecallTest(PrecisionRecall):  # noqa: D101
    attributes: dict = {"data_source": "test"}


class PredictionError(Performance):  # noqa: D101
    accessor: ClassVar[str] = "metrics.prediction_error"
    key: str = "prediction_error"
    verbose_name: str = "Prediction error"


class PredictionErrorTrain(PredictionError):  # noqa: D101
    attributes: dict = {"data_source": "train"}


class PredictionErrorTest(PredictionError):  # noqa: D101
    attributes: dict = {"data_source": "test"}


class Roc(Performance):  # noqa: D101
    accessor: ClassVar[str] = "metrics.roc"
    key: str = "roc"
    verbose_name: str = "ROC"


class RocTrain(Roc):  # noqa: D101
    attributes: dict = {"data_source": "train"}


class RocTest(Roc):  # noqa: D101
    attributes: dict = {"data_source": "test"}



"""Class definition of the payload used to send an accuracy metric to ``hub``."""

from __future__ import annotations

from typing import ClassVar, Literal

from .metric import CrossValidationReportMetric, EstimatorReportMetric


class Accuracy(EstimatorReportMetric):  # noqa: D101
    accessor: ClassVar[Literal["metrics.accuracy"]] = "metrics.accuracy"
    name: Literal["accuracy"] = "accuracy"
    verbose_name: Literal["Accuracy"] = "Accuracy"
    greater_is_better: Literal[True] = True


class AccuracyTrain(Accuracy):  # noqa: D101
    data_source: Literal["train"] = "train"


class AccuracyTest(Accuracy):  # noqa: D101
    data_source: Literal["test"] = "test"


class AccuracyMean(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[Literal["metrics.accuracy"]] = "metrics.accuracy"
    aggregate: ClassVar[Literal["mean"]] = "mean"
    name: Literal["accuracy_mean"] = "accuracy_mean"
    verbose_name: Literal["Accuracy - MEAN"] = "Accuracy - MEAN"
    greater_is_better: Literal[True] = True


class AccuracyTrainMean(AccuracyMean):  # noqa: D101
    data_source: Literal["train"] = "train"


class AccuracyTestMean(AccuracyMean):  # noqa: D101
    data_source: Literal["test"] = "test"


class AccuracyStd(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[Literal["metrics.accuracy"]] = "metrics.accuracy"
    aggregate: ClassVar[Literal["std"]] = "std"
    name: Literal["accuracy_std"] = "accuracy_std"
    verbose_name: Literal["Accuracy - STD"] = "Accuracy - STD"
    greater_is_better: Literal[False] = False


class AccuracyTrainStd(AccuracyStd):  # noqa: D101
    data_source: Literal["train"] = "train"


class AccuracyTestStd(AccuracyStd):  # noqa: D101
    data_source: Literal["test"] = "test"



"""Class definition of the payload used to send a data category media to ``hub``."""

from functools import cached_property
from inspect import signature
from typing import Literal

import numpy as np
from pydantic import Field, computed_field

from skore_hub_project import switch_mpl_backend
from skore_hub_project.protocol import CrossValidationReport, EstimatorReport

from .media import Media, Representation


def _to_native(obj):
    """Walk an object and cast all numpy types to native type.

    Useful for json serialization.
    """
    if isinstance(obj, dict):
        return {k: _to_native(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_to_native(v) for v in obj]
    elif isinstance(obj, tuple):
        return tuple(_to_native(v) for v in obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, np.generic):
        return obj.item()
    else:
        return obj


class TableReport(Media):  # noqa: D101
    report: EstimatorReport | CrossValidationReport = Field(repr=False, exclude=True)
    key: str = "table_report"
    verbose_name: str = "Table report"
    category: Literal["data"] = "data"

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def representation(self) -> Representation:  # noqa: D102
        with switch_mpl_backend():
            function = self.report.data.analyze
            function_parameters = signature(function).parameters
            function_kwargs = {
                k: v for k, v in self.attributes.items() if k in function_parameters
            }

            table_report_display = function(**function_kwargs)
            table_report = table_report_display.summary

            table_report["extract_head"] = (
                table_report["dataframe"].head(3).to_dict(orient="split")
            )

            table_report["extract_tail"] = (
                table_report["dataframe"].tail(3).to_dict(orient="split")
            )

            del table_report["dataframe"]
            del table_report["sample_table"]

            return Representation(
                media_type="application/vnd.skrub.table-report.v1+json",
                value=_to_native(table_report),
            )


class TableReportTrain(TableReport):  # noqa: D101
    attributes: dict = {"data_source": "train"}


class TableReportTest(TableReport):  # noqa: D101
    attributes: dict = {"data_source": "test"}



"""Class definitions of the payloads used to send a metric to ``hub``."""

from .accuracy import (
    AccuracyTest,
    AccuracyTestMean,
    AccuracyTestStd,
    AccuracyTrain,
    AccuracyTrainMean,
    AccuracyTrainStd,
)
from .brier_score import (
    BrierScoreTest,
    BrierScoreTestMean,
    BrierScoreTestStd,
    BrierScoreTrain,
    BrierScoreTrainMean,
    BrierScoreTrainStd,
)
from .log_loss import (
    LogLossTest,
    LogLossTestMean,
    LogLossTestStd,
    LogLossTrain,
    LogLossTrainMean,
    LogLossTrainStd,
)
from .precision import (
    PrecisionTest,
    PrecisionTestMean,
    PrecisionTestStd,
    PrecisionTrain,
    PrecisionTrainMean,
    PrecisionTrainStd,
)
from .r2 import (
    R2Test,
    R2TestMean,
    R2TestStd,
    R2Train,
    R2TrainMean,
    R2TrainStd,
)
from .recall import (
    RecallTest,
    RecallTestMean,
    RecallTestStd,
    RecallTrain,
    RecallTrainMean,
    RecallTrainStd,
)
from .rmse import (
    RmseTest,
    RmseTestMean,
    RmseTestStd,
    RmseTrain,
    RmseTrainMean,
    RmseTrainStd,
)
from .roc_auc import (
    RocAucTest,
    RocAucTestMean,
    RocAucTestStd,
    RocAucTrain,
    RocAucTrainMean,
    RocAucTrainStd,
)
from .timing import (
    FitTime,
    FitTimeMean,
    FitTimeStd,
    PredictTimeTest,
    PredictTimeTestMean,
    PredictTimeTestStd,
    PredictTimeTrain,
    PredictTimeTrainMean,
    PredictTimeTrainStd,
)

__all__ = [
    "AccuracyTest",
    "AccuracyTestMean",
    "AccuracyTestStd",
    "AccuracyTrain",
    "AccuracyTrainMean",
    "AccuracyTrainStd",
    "BrierScoreTest",
    "BrierScoreTestMean",
    "BrierScoreTestStd",
    "BrierScoreTrain",
    "BrierScoreTrainMean",
    "BrierScoreTrainStd",
    "FitTime",
    "FitTimeMean",
    "FitTimeStd",
    "LogLossTest",
    "LogLossTestMean",
    "LogLossTestStd",
    "LogLossTrain",
    "LogLossTrainMean",
    "LogLossTrainStd",
    "PrecisionTest",
    "PrecisionTestMean",
    "PrecisionTestStd",
    "PrecisionTrain",
    "PrecisionTrainMean",
    "PrecisionTrainStd",
    "PredictTimeTest",
    "PredictTimeTestMean",
    "PredictTimeTestStd",
    "PredictTimeTrain",
    "PredictTimeTrainMean",
    "PredictTimeTrainStd",
    "R2Test",
    "R2TestMean",
    "R2TestStd",
    "R2Train",
    "R2TrainMean",
    "R2TrainStd",
    "RecallTest",
    "RecallTestMean",
    "RecallTestStd",
    "RecallTrain",
    "RecallTrainMean",
    "RecallTrainStd",
    "RmseTest",
    "RmseTestMean",
    "RmseTestStd",
    "RmseTrain",
    "RmseTrainMean",
    "RmseTrainStd",
    "RocAucTest",
    "RocAucTestMean",
    "RocAucTestStd",
    "RocAucTrain",
    "RocAucTrainMean",
    "RocAucTrainStd",
]



"""Class definition of the payload used to send a R² metric to ``hub``."""

from __future__ import annotations

from typing import ClassVar, Literal

from .metric import CrossValidationReportMetric, EstimatorReportMetric


class R2(EstimatorReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.r2"
    name: str = "r2"
    verbose_name: str = "R²"
    greater_is_better: bool = True


class R2Train(R2):  # noqa: D101
    data_source: Literal["train"] = "train"


class R2Test(R2):  # noqa: D101
    data_source: Literal["test"] = "test"


class R2Mean(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.r2"
    aggregate: ClassVar[Literal["mean"]] = "mean"
    name: str = "r2_mean"
    verbose_name: str = "R² - MEAN"
    greater_is_better: bool = True


class R2TrainMean(R2Mean):  # noqa: D101
    data_source: Literal["train"] = "train"


class R2TestMean(R2Mean):  # noqa: D101
    data_source: Literal["test"] = "test"


class R2Std(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.r2"
    aggregate: ClassVar[Literal["std"]] = "std"
    name: str = "r2_std"
    verbose_name: str = "R² - STD"
    greater_is_better: bool = False


class R2TrainStd(R2Std):  # noqa: D101
    data_source: Literal["train"] = "train"


class R2TestStd(R2Std):  # noqa: D101
    data_source: Literal["test"] = "test"



"""Token used for ``skore hub`` authentication."""

from __future__ import annotations

import json
import pathlib
import tempfile
from datetime import datetime, timezone

from ..client.api import post_oauth_refresh_token


def filepath():
    """Filepath used to save the tokens on disk."""
    return pathlib.Path(tempfile.gettempdir(), "skore.token")


def save(access: str, refreshment: str, expires_at: str):
    """
    Save the token.

    Save the tokens to the disk to prevent user to login more than once, as long as
    the access token is valid or can be refreshed.
    """
    filepath().write_text(
        json.dumps(
            (
                access,
                refreshment,
                expires_at,
            )
        )
    )


def exists() -> bool:
    """Existence of the token."""
    return filepath().exists()


def access(*, refresh=True) -> str:
    """
    Access token.

    Parameters
    ----------
    refresh : bool, optional
        Refresh the token on-the-fly if necessary, default True.
    """
    access, refreshment, expiration = json.loads(filepath().read_text())

    if refresh and datetime.fromisoformat(expiration) <= datetime.now(timezone.utc):
        # Retrieve freshly updated tokens
        access, refreshment, expiration = post_oauth_refresh_token(refreshment)

        # Re-save the refreshed tokens
        save(access, refreshment, expiration)

    return access



"""Class definition of the payload used to upload and send an artefact to ``hub``."""

from __future__ import annotations

from abc import ABC
from functools import cached_property
from typing import Any

from pydantic import BaseModel, ConfigDict, Field, computed_field

from skore_hub_project import Project
from skore_hub_project.artefact.upload import upload
from skore_hub_project.protocol import CrossValidationReport, EstimatorReport


class Artefact(ABC, BaseModel):
    """Payload used to send the artefact of a report to ``hub``."""

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    project: Project = Field(repr=False, exclude=True)
    object: Any = Field(repr=False, exclude=True)


class EstimatorReportArtefact(Artefact):
    """
    Payload used to upload and send an estimator report artefact to ``hub``.

    Notes
    -----
    It uploads the report to artefacts storage in a lazy way.

    The report is uploaded without its cache, to avoid salting the checksum.
    The report is primarily pickled on disk to reduce RAM footprint.
    """

    object: EstimatorReport = Field(repr=False, exclude=True, alias="report")

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def checksum(self) -> str:
        """Checksum, useful for retrieving the artefact from artefact storage."""
        cache = self.object._cache
        self.object._cache = {}

        try:
            return upload(self.project, self.object, "estimator-report")
        finally:
            self.object._cache = cache


class CrossValidationReportArtefact(Artefact):
    """
    Payload used to upload and send a cross-validation report artefact to ``hub``.

    Notes
    -----
    It uploads the report to artefacts storage in a lazy way.

    The report is uploaded without its cache, to avoid salting the checksum.
    The report is primarily pickled on disk to reduce RAM footprint.
    """

    object: CrossValidationReport = Field(repr=False, exclude=True, alias="report")

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def checksum(self) -> str:
        """Checksum, useful for retrieving the artefact from artefact storage."""
        reports = [self.object] + self.object.estimator_reports_
        caches = []

        for report in reports:
            caches.append(report._cache)
            report._cache = {}

        try:
            return upload(self.project, self.object, "cross-validation-report")
        finally:
            for i, report in enumerate(reports):
                report._cache = caches[i]



"""Class definition of the payload used to send a timing metric to ``hub``."""

from __future__ import annotations

from functools import cached_property
from typing import ClassVar, Literal

from pandas import DataFrame, Series
from pydantic import Field, computed_field

from skore_hub_project.protocol import CrossValidationReport, EstimatorReport

from .metric import Metric, cast_to_float


class FitTime(Metric):  # noqa: D101
    report: EstimatorReport = Field(repr=False, exclude=True)
    name: str = "fit_time"
    verbose_name: str = "Fit time (s)"
    greater_is_better: bool = False
    position: int = 1

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def value(self) -> float | None:  # noqa: D102
        timings: dict = self.report.metrics.timings()
        fit_time = timings.get("fit_time")

        return cast_to_float(fit_time)


class FitTimeAggregate(Metric):  # noqa: D101
    # ``report.metrics.timings()``
    #
    #                             mean       std
    # Fit time (s)                 ...       ...
    # Predict time test (s)        ...       ...
    # Predict time train (s)       ...       ...

    report: CrossValidationReport = Field(repr=False, exclude=True)
    aggregate: ClassVar[Literal["mean", "std"]]
    greater_is_better: bool = False

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def value(self) -> float | None:  # noqa: D102
        timings: DataFrame = self.report.metrics.timings(aggregate=self.aggregate)

        try:
            fit_times: Series = timings.loc["Fit time (s)"]
        except KeyError:
            return None

        return cast_to_float(fit_times.iloc[0])


class FitTimeMean(FitTimeAggregate):  # noqa: D101
    aggregate: ClassVar[Literal["mean"]] = "mean"
    name: str = "fit_time_mean"
    verbose_name: str = "Fit time (s) - MEAN"
    position: int = 1


class FitTimeStd(FitTimeAggregate):  # noqa: D101
    aggregate: ClassVar[Literal["std"]] = "std"
    name: str = "fit_time_std"
    verbose_name: str = "Fit time (s) - STD"


class PredictTime(Metric):  # noqa: D101
    report: EstimatorReport = Field(repr=False, exclude=True)
    name: str = "predict_time"
    verbose_name: str = "Predict time (s)"
    greater_is_better: bool = False
    position: int = 2

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def value(self) -> float | None:  # noqa: D102
        timings: dict = self.report.metrics.timings()
        predict_time = timings.get(f"predict_time_{self.data_source}")

        return cast_to_float(predict_time)


class PredictTimeTrain(PredictTime):  # noqa: D101
    data_source: Literal["train"] = "train"


class PredictTimeTest(PredictTime):  # noqa: D101
    data_source: Literal["test"] = "test"


class PredictTimeAggregate(Metric):  # noqa: D101
    # ``report.metrics.timings()``
    #
    #                             mean       std
    # Fit time (s)                 ...       ...
    # Predict time test (s)        ...       ...
    # Predict time train (s)       ...       ...

    report: CrossValidationReport = Field(repr=False, exclude=True)
    aggregate: ClassVar[Literal["mean", "std"]]
    greater_is_better: bool = False

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def value(self) -> float | None:  # noqa: D102
        timings: DataFrame = self.report.metrics.timings(aggregate=self.aggregate)

        try:
            predict_times: Series = timings.loc[f"Predict time {self.data_source} (s)"]
        except KeyError:
            return None

        return cast_to_float(predict_times.iloc[0])


class PredictTimeMean(PredictTimeAggregate):  # noqa: D101
    aggregate: ClassVar[Literal["mean"]] = "mean"
    name: str = "predict_time_mean"
    verbose_name: str = "Predict time (s) - MEAN"
    position: int = 2


class PredictTimeTrainMean(PredictTimeMean):  # noqa: D101
    data_source: Literal["train"] = "train"


class PredictTimeTestMean(PredictTimeMean):  # noqa: D101
    data_source: Literal["test"] = "test"


class PredictTimeStd(PredictTimeAggregate):  # noqa: D101
    aggregate: ClassVar[Literal["std"]] = "std"
    name: str = "predict_time_std"
    verbose_name: str = "Predict time (s) - STD"


class PredictTimeTrainStd(PredictTimeStd):  # noqa: D101
    data_source: Literal["train"] = "train"


class PredictTimeTestStd(PredictTimeStd):  # noqa: D101
    data_source: Literal["test"] = "test"



"""Protocols definition used to remove adherence to ``skore`` for type checking."""

from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class EstimatorReport(Protocol):
    """Protocol equivalent to ``skore.EstimatorReport``."""

    _cache: Any
    metrics: Any
    data: Any
    ml_task: Any
    estimator: Any
    estimator_name_: Any
    X_train: Any
    y_train: Any
    X_test: Any
    y_test: Any
    fit: Any


@runtime_checkable
class CrossValidationReport(Protocol):
    """Protocol equivalent to ``skore.CrossValidationReport``."""

    _cache: Any
    metrics: Any
    data: Any
    estimator_reports_: Any
    ml_task: Any
    estimator: Any
    estimator_name_: Any
    X: Any
    y: Any
    splitter: Any
    split_indices: Any



"""Class definition of the payload used to send a cross-validation report to ``hub``."""

from collections import defaultdict
from functools import cached_property
from typing import ClassVar, cast

import numpy as np
from pydantic import Field, computed_field
from sklearn.model_selection import BaseCrossValidator
from sklearn.model_selection._split import _CVIterableWrapper

from skore_hub_project.artefact import CrossValidationReportArtefact
from skore_hub_project.media import EstimatorHtmlRepr
from skore_hub_project.media.data import TableReport
from skore_hub_project.media.media import Media
from skore_hub_project.metric import (
    AccuracyTestMean,
    AccuracyTestStd,
    AccuracyTrainMean,
    AccuracyTrainStd,
    BrierScoreTestMean,
    BrierScoreTestStd,
    BrierScoreTrainMean,
    BrierScoreTrainStd,
    FitTimeMean,
    FitTimeStd,
    LogLossTestMean,
    LogLossTestStd,
    LogLossTrainMean,
    LogLossTrainStd,
    PrecisionTestMean,
    PrecisionTestStd,
    PrecisionTrainMean,
    PrecisionTrainStd,
    PredictTimeTestMean,
    PredictTimeTestStd,
    PredictTimeTrainMean,
    PredictTimeTrainStd,
    R2TestMean,
    R2TestStd,
    R2TrainMean,
    R2TrainStd,
    RecallTestMean,
    RecallTestStd,
    RecallTrainMean,
    RecallTrainStd,
    RmseTestMean,
    RmseTestStd,
    RmseTrainMean,
    RmseTrainStd,
    RocAucTestMean,
    RocAucTestStd,
    RocAucTrainMean,
    RocAucTrainStd,
)
from skore_hub_project.metric.metric import Metric
from skore_hub_project.protocol import CrossValidationReport
from skore_hub_project.report.estimator_report import EstimatorReportPayload
from skore_hub_project.report.report import ReportPayload


class CrossValidationReportPayload(ReportPayload):
    """
    Payload used to send a cross-validation report to ``hub``.

    Attributes
    ----------
    METRICS : ClassVar[tuple[Metric, ...]]
        The metric classes that have to be computed from the report.
    MEDIAS : ClassVar[tuple[Media, ...]]
        The media classes that have to be computed from the report.
    project : Project
        The project to which the report payload should be sent.
    report : CrossValidationReport
        The report on which to calculate the payload to be sent.
    upload : bool, optional
        Upload the report to the artefacts storage, default True.
    key : str
        The key to associate to the report.
    """

    METRICS: ClassVar[tuple[Metric, ...]] = cast(
        tuple[Metric, ...],
        (
            AccuracyTestMean,
            AccuracyTestStd,
            AccuracyTrainMean,
            AccuracyTrainStd,
            BrierScoreTestMean,
            BrierScoreTestStd,
            BrierScoreTrainMean,
            BrierScoreTrainStd,
            LogLossTestMean,
            LogLossTestStd,
            LogLossTrainMean,
            LogLossTrainStd,
            PrecisionTestMean,
            PrecisionTestStd,
            PrecisionTrainMean,
            PrecisionTrainStd,
            R2TestMean,
            R2TestStd,
            R2TrainMean,
            R2TrainStd,
            RecallTestMean,
            RecallTestStd,
            RecallTrainMean,
            RecallTrainStd,
            RmseTestMean,
            RmseTestStd,
            RmseTrainMean,
            RmseTrainStd,
            RocAucTestMean,
            RocAucTestStd,
            RocAucTrainMean,
            RocAucTrainStd,
            # timings must be calculated last
            FitTimeMean,
            FitTimeStd,
            PredictTimeTestMean,
            PredictTimeTestStd,
            PredictTimeTrainMean,
            PredictTimeTrainStd,
        ),
    )
    MEDIAS: ClassVar[tuple[Media, ...]] = cast(
        tuple[Media, ...],
        (
            EstimatorHtmlRepr,
            TableReport,
        ),
    )

    report: CrossValidationReport = Field(repr=False, exclude=True)

    def model_post_init(self, context):  # noqa: D102
        if "classification" in self.ml_task:
            class_to_class_indice = defaultdict(lambda: len(class_to_class_indice))

            self.__sample_to_class_index = [
                class_to_class_indice[sample] for sample in self.report.y
            ]

            assert len(self.__sample_to_class_index) == len(self.report.X)

            self.__classes = [str(class_) for class_ in class_to_class_indice]

            assert max(self.__sample_to_class_index) == (len(self.__classes) - 1)
        else:
            self.__sample_to_class_index = None
            self.__classes = None

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def dataset_size(self) -> int:
        """Size of the dataset."""
        return len(self.report.X)

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def splitting_strategy_name(self) -> str:
        """The name of the splitting strategy used by the report."""
        is_sklearn_splitter = isinstance(self.report.splitter, BaseCrossValidator)
        is_iterable_splitter = isinstance(self.report.splitter, _CVIterableWrapper)
        is_standard_strategy = is_sklearn_splitter and (not is_iterable_splitter)

        return (
            is_standard_strategy and self.report.splitter.__class__.__name__ or "custom"
        )

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def splits(self) -> list[list[float]]:
        """
        Distribution between train and test by split.

        The distribution of each split is computed by dividing the split into a maximum
        of 200 buckets, and averaging the number of samples belonging to the test-set in
        each of these buckets.
        """
        distributions = []
        buckets_number = min(len(self.report.X), 200)

        for _, test_indices in self.report.split_indices:
            split = np.zeros(len(self.report.X), dtype=int)
            split[test_indices] = 1

            distributions.append(
                [
                    float(np.mean(bucket))
                    for bucket in np.array_split(split, buckets_number)
                ]
            )

        return distributions

    groups: list[int] | None = None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def class_names(self) -> list[str] | None:
        """In classification, the class names of the dataset used in the report."""
        return self.__classes

    @computed_field  # type: ignore[prop-decorator]
    @property
    def classes(self) -> list[int] | None:
        """
        In classification, the distribution of the classes in the dataset.

        The distribution is computed by dividing the dataset into a maximum of 200
        buckets, and noting the dominant class in each of these buckets.
        """
        if self.__sample_to_class_index is None:
            return None

        buckets_number = min(len(self.__sample_to_class_index), 200)
        buckets = np.array_split(self.__sample_to_class_index, buckets_number)

        return [int(np.bincount(bucket).argmax()) for bucket in buckets]

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def estimators(self) -> list[EstimatorReportPayload]:
        """The estimators used in each split by the report in a payload format."""
        return [
            EstimatorReportPayload(
                project=self.project,
                report=report,
                upload=False,
                key=f"{self.key}:estimator-report",
            )
            for report in self.report.estimator_reports_
        ]

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def parameters(self) -> CrossValidationReportArtefact | dict[()]:
        """
        The checksum of the instance.

        The checksum of the instance that was assigned after being uploaded to the
        artefact storage. It is based on its ``joblib`` serialization and mainly used to
        retrieve it from the artefacts storage.

        .. deprecated
          The ``parameters`` property will be removed in favor of a new ``checksum``
          property in a near future.
        """
        if self.upload:
            return CrossValidationReportArtefact(
                project=self.project,
                report=self.report,
            )
        return {}



"""Class definition of the payload used to send a media to ``hub``."""

from abc import ABC, abstractmethod
from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, computed_field


class Representation(BaseModel):
    """Payload used to send the representation of a media to ``hub``."""

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    media_type: str
    value: Any


class Media(ABC, BaseModel):
    """
    Payload used to send a media to ``hub``.

    Attributes
    ----------
    key : str
        Key/name of the media.
    verbose_name : str
        Verbose name of the media.
    category: Literal["performance", "feature_importance", "model", "data"]
        Category of the media.
    attributes: dict, optional
        Attributes of the media usually used to customize the calculation function,
        default empty dict.
    parameters: dict
        .. deprecated
          The ``parameters`` property is unused and will be removed.
    """

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    key: str
    verbose_name: str
    category: Literal["performance", "feature_importance", "model", "data"]
    attributes: dict = {}
    parameters: dict = {}

    @computed_field  # type: ignore[prop-decorator]
    @property
    @abstractmethod
    def representation(self) -> Representation | None:
        """The representation of the media."""



"""Class definition of the payload used to send a precision metric to ``hub``."""

from __future__ import annotations

from functools import cached_property
from typing import ClassVar, Literal

from pydantic import computed_field

from .metric import CrossValidationReportMetric, EstimatorReportMetric, cast_to_float


class Precision(EstimatorReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.precision"
    name: str = "precision"
    verbose_name: str = "Precision (macro)"
    greater_is_better: bool = True

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def value(self) -> float | None:  # noqa: D102
        try:
            function = self.report.metrics.precision
        except AttributeError:
            return None

        return cast_to_float(function(data_source=self.data_source, average="macro"))


class PrecisionTrain(Precision):  # noqa: D101
    data_source: Literal["train"] = "train"


class PrecisionTest(Precision):  # noqa: D101
    data_source: Literal["test"] = "test"


class PrecisionMean(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.precision"
    aggregate: ClassVar[Literal["mean"]] = "mean"
    name: str = "precision_mean"
    verbose_name: str = "Precision (macro) - MEAN"
    greater_is_better: bool = True

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def value(self) -> float | None:  # noqa: D102
        try:
            function = self.report.metrics.precision
        except AttributeError:
            return None

        dataframe = function(
            data_source=self.data_source, aggregate=self.aggregate, average="macro"
        )

        return cast_to_float(dataframe.iloc[0, 0])


class PrecisionTrainMean(PrecisionMean):  # noqa: D101
    data_source: Literal["train"] = "train"


class PrecisionTestMean(PrecisionMean):  # noqa: D101
    data_source: Literal["test"] = "test"


class PrecisionStd(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.precision"
    aggregate: ClassVar[Literal["std"]] = "std"
    name: str = "precision_std"
    verbose_name: str = "Precision (macro) - STD"
    greater_is_better: bool = False

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def value(self) -> float | None:  # noqa: D102
        try:
            function = self.report.metrics.precision
        except AttributeError:
            return None

        dataframe = function(
            data_source=self.data_source, aggregate=self.aggregate, average="macro"
        )

        return cast_to_float(dataframe.iloc[0, 0])


class PrecisionTrainStd(PrecisionStd):  # noqa: D101
    data_source: Literal["train"] = "train"


class PrecisionTestStd(PrecisionStd):  # noqa: D101
    data_source: Literal["test"] = "test"



"""Logout from ``skore hub``."""

from __future__ import annotations

from . import token as Token


def logout():
    """Logout from ``skore hub`` by deleting tokens."""
    if Token.exists():
        Token.filepath().unlink()
